"""This private module contains functions related to specific calculations within
pyCIAM that are called by the ``run`` module."""

import numpy as np
import xarray as xr
from scipy.stats import gumbel_r

from .damage_funcs import diaz_ddf_i, diaz_dmf_i


def _get_surge_heights_probs(
    min_surge_ht, max_surge_ht, gumbel_params, n_surge_heights=100
):
    """Create an array of ``n_surge_heights`` surge heights and associated probabilities
    to apply in CIAM in order to sample an appropriate range of plausible surge heights.
    """

    # get gumbel params
    loc = gumbel_params.sel(params="loc", drop=True)
    scale = gumbel_params.sel(params="scale", drop=True)

    # get array of surge heights to model
    surge_hts = xr.apply_ufunc(
        np.linspace,
        min_surge_ht,
        max_surge_ht,
        n_surge_heights,
        output_core_dims=[["surge_height"]],
        kwargs={"axis": -1},
        dask="parallelized",
        dask_gufunc_kwargs={"output_sizes": {"surge_height": n_surge_heights}},
    )

    # get probability of falling into each surge height bin
    cdf = xr.apply_ufunc(
        gumbel_r.cdf,
        surge_hts,
        kwargs={
            "loc": loc.broadcast_like(surge_hts),
            "scale": scale.broadcast_like(surge_hts),
        },
        dask="parallelized",
    )
    pdf = cdf.diff(dim="surge_height", label="lower")

    # get the midpoint surge height of each bin
    surge_hts_mid = ((surge_hts + surge_hts.shift(surge_height=-1)) / 2).sel(
        surge_height=pdf.surge_height.values
    )

    # add in the final 1/10000th of the cdf and conservatively assign it the 1:10000 yr
    # height
    last_surge = surge_hts_mid.isel(surge_height=0, drop=True).copy()
    last_surge.data = np.ones_like(last_surge) * max_surge_ht
    surge_hts_mid = xr.concat((surge_hts_mid, last_surge), dim="surge_height")
    pdf = xr.concat((pdf, 1 - cdf.isel(surge_height=-1)), dim="surge_height")

    return xr.Dataset({"surge": surge_hts_mid, "p": pdf})


def _calc_storm_capital_and_mortality(
    esl,
    exp_dens,
    bin_bounds,
    R,
    template,
    ddf_i=diaz_ddf_i,
    dmf_i=diaz_dmf_i,
    ddf_kwargs={},
    dmf_kwargs={},
):
    min_ht = np.maximum(bin_bounds.sel(bound="lower", drop=True), R)
    max_ht = np.maximum(bin_bounds.sel(bound="upper", drop=True), R)
    depth_st = np.maximum(esl - max_ht, 0)
    depth_end = np.maximum(esl - min_ht, 0)
    del min_ht, max_ht

    storm_capital = (
        (ddf_i(depth_st, depth_end, **ddf_kwargs) * exp_dens.K)
        .where(depth_end > 0)
        .sum("elev")
    )
    storm_mortality = (
        (dmf_i(depth_st, depth_end, **dmf_kwargs) * exp_dens.pop)
        .where(depth_end > 0)
        .sum("elev")
    )

    return (
        xr.Dataset({"stormCapital": storm_capital, "stormPopulation": storm_mortality})
        .unstack()
        .reindex_like(template, fill_value=0)
    )


def _calc_storm_damages_no_resilience(
    esl,
    exposure,
    elev_bounds,
    R,
    H,
    ddf_i,
    dmf_i,
    ddf_kwargs={},
    dmf_kwargs={},
    stack=False,
    surge_probs=None,
):
    """Estimate storm-related damages before applying resilience factor."""

    elev_bin_widths = elev_bounds.sel(bound="upper", drop=True) - elev_bounds.sel(
        bound="lower", drop=True
    )

    # turn exposure into exposure density
    this_exposure_dens = exposure / elev_bin_widths

    valid = esl >= H
    esl = esl.where(valid)
    if stack:
        esl = esl.stack(stacked=esl.dims).dropna("stacked")
        if isinstance(R, xr.DataArray):
            R = R.where(valid)
            R = R.stack(stacked=esl.stacked.to_index().names).sel(stacked=esl.stacked)

    if surge_probs is None:
        out = _calc_storm_capital_and_mortality(
            esl,
            this_exposure_dens,
            elev_bounds,
            R,
            valid,
            ddf_i=ddf_i,
            dmf_i=dmf_i,
            ddf_kwargs=ddf_kwargs,
            dmf_kwargs=dmf_kwargs,
        )
    else:
        init_da = xr.zeros_like(valid)
        if "surge_height" in init_da.dims:
            init_da = init_da.sel
        out = None
        for h in surge_probs.surge_height.values:
            this_surge_probs = surge_probs.sel(surge_height=h, drop=True)
            if this_surge_probs.sum() == 0:
                continue

            this_esl = esl.sel(surge_height=h, drop=True)
            if isinstance(R, xr.DataArray) and "stacked" in R.dims:
                this_R = R.sel(surge_height=h, drop=True)
            else:
                this_R = R

            this_out = (
                _calc_storm_capital_and_mortality(
                    this_esl,
                    this_exposure_dens,
                    elev_bounds,
                    this_R,
                    valid.sel(surge_height=h, drop=True),
                    ddf_i=ddf_i,
                    dmf_i=dmf_i,
                    ddf_kwargs=ddf_kwargs,
                    dmf_kwargs=dmf_kwargs,
                )
                * this_surge_probs
            )
            if out is None:
                out = this_out
            else:
                out += this_out
    return out



import os
import zipfile
from pathlib import Path

import numpy as np
import xarray as xr
from dask.utils import tmpfile
from gcsfs import GCSFileSystem

from pyCIAM import __file__

FS = GCSFileSystem(token="/opt/gcsfuse_tokens/rhg-data.json")

SLIIDERS_VERS = "v1.0"
RES_VERS = "v1.0"

# quantiles of local SLR at which to run analysis
QUANTILES = [0.05, 0.5, 0.95]

# Output dataset attrs
HISTORY = """version 1.0: Version associated with Depsky et al. 2022"""
AUTHOR = "Ian Bolliger"
CONTACT = "ian.bolliger@blackrock.com"

# Filepaths
DIR_HOME = Path("/gcs/rhg-data/impactlab-rhg/coastal")

PATH_PARAMS = Path(__file__).parents[1] / "params.json"
PATH_PARAMS_DIAZ = Path(__file__).parents[1] / "params_diaz.json"

DIR_SLIIDERS = DIR_HOME / "sliiders"
DIR_SLIIDERS_OUTPUT = DIR_SLIIDERS / "output"
PATH_SLIIDERS_ECON = FS.get_mapper(
    DIR_SLIIDERS_OUTPUT.relative_to("/gcs") / f"sliiders-econ-{SLIIDERS_VERS}.zarr"
)
PATH_SLIIDERS_SLR = FS.get_mapper(
    DIR_SLIIDERS_OUTPUT.relative_to("/gcs") / f"sliiders-slr-{SLIIDERS_VERS}.zarr"
)

DIR_CIAM = DIR_HOME / "ciam_paper"
DIR_DATA = DIR_CIAM / "data"
DIR_RAW = DIR_DATA / "raw"
DIR_INT = DIR_DATA / "int"
DIR_RES = DIR_CIAM / "results"
DIR_SHP = DIR_HOME / "data" / "raw" / "ciam_inputs" / "shapefiles"

PATH_SLIIDERS_SLR_QUANTILES = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / f"sliiders-slr-{SLIIDERS_VERS}-quantiles.zarr"
)
PATH_SLIIDERS_ECON_SEG = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / f"sliiders-econ-{SLIIDERS_VERS}-seg.zarr"
)

PATH_SURGE_LOOKUP_SEG = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / f"surge_lookup_seg_{SLIIDERS_VERS}.zarr"
)

PATH_SURGE_LOOKUP = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / f"surge_lookup_{SLIIDERS_VERS}.zarr"
)
PATH_REFA = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / f"refA_by_movefactor_{SLIIDERS_VERS}.zarr"
)

PATH_DIAZ_INPUTS_RAW = FS.get_mapper(
    DIR_SLIIDERS.relative_to("/gcs") / "raw" / "CIAM_2016" / "diaz2016_inputs_raw.zarr"
)
PATH_DIAZ_INPUTS_INT = FS.get_mapper(
    DIR_INT.relative_to("/gcs") / "diaz2016_inputs.zarr"
)

PATH_MOVEFACTOR_DATA = FS.get_mapper(
    DIR_RES.relative_to("/gcs")
    / f"suboptimal_capital_by_movefactor_{SLIIDERS_VERS}.zarr"
)

PATH_QUANTILE_RES = FS.get_mapper(
    DIR_RES.relative_to("/gcs") / "pyCIAM_results_quantiles.zarr"
)
PATH_DIAZ_RES = FS.get_mapper(DIR_RES.relative_to("/gcs") / "pyCIAM_results_diaz.zarr")

PATH_PWT_RAW = DIR_SLIIDERS / "raw" / "exposure" / "ypk" / "pwt_100.xlsx"

PATH_BORDERS = (
    DIR_SHP
    / "ne_50m_admin_0_boundary_lines_land"
    / "ne_50m_admin_0_boundary_lines_land.shp"
)
PATH_COASTLINES = DIR_SHP / "ne_110m_coastline" / "ne_110m_coastline.shp"
DIR_GADM = DIR_SHP / "gadm"
PATH_GADM_0 = DIR_GADM / "gadm36_0.shp"
PATH_GADM_1 = DIR_GADM / "gadm36_1.shp"


def get_ncc_slr(inputs, seg_var="seg_adm"):
    site_ids = inputs.SLR_site_id.values
    segs = inputs[seg_var].values
    slr = (
        xr.open_zarr(PATH_SLIIDERS_SLR, chunks=None)
        .lsl_ncc_msl00.sel(site_id=site_ids, drop=True)
        .quantile(QUANTILES, dim="mc_sample_id")
        .rename(site_id="seg")
    )
    slr["seg"] = segs
    slr = slr.reindex(
        year=np.concatenate(([2000], slr.year.values)), fill_value=0
    ).interp(year=inputs.year.values)
    return slr


def upload_pkg(client, pkg_dir):
    with tmpfile(extension="zip") as f:
        zipf = zipfile.ZipFile(f, "w", zipfile.ZIP_DEFLATED)
        for root, dirs, files in os.walk(pkg_dir):
            for file in files:
                zipf.write(
                    os.path.join(root, file),
                    os.path.relpath(
                        os.path.join(root, file), os.path.join(pkg_dir, "..")
                    ),
                )
        zipf.close()
        client.upload_file(f)



"""This module contains functions related to loading and saving inputs and intermediate
outputs used in running pyCIAM.

Functions
    prep_sliiders
    load_ciam_inputs
    load_diaz_inputs
"""

import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from ._utils import _s2d


def prep_sliiders(input_store, seg_vals, constants={}, seg_var="seg_adm", selectors={}):
    """Import the SLIIDERS-ECON dataset (or a different dataset formatted analogously),
    format, and calculate derived variables so that it can be used by the functions that
    implement pyCIAM.

    Parameters
    ----------
    input_store : Path-like
        Path to zarr store that contains the SLIIDERS-ECON dataset (or one formatted
        similarly. May be a :py:class:`fsspec.FSMap` object.
    seg_vals : list of str
        Defines the subset of regions (along dimension `seg_var`) that the function
        will prep. Subsets are used to run CIAM in parallel.
    constants : dict, default {}
        Defines a list of constants, typically contained in a JSON file, which will be
        merged as new variables into the :py:class:`xarray.Dataset` along with the
        processed variables from `input_store`.
    seg_var : str, default "seg_var"
        The name of the dimension in `input_store` along which the function will
        subset using `seg_vals`
    selectors : dict
        Defines additional dimensions and values with which to subset `input_store`.
        For example, the `ssp` dimension in SLIIDERS-ECON defines different
        Shared Socioeconomic Pathways (SSPs). If you only wish to proocess SSP2, you
        could specify ``selectors={"ssp": "SSP2"}``.

    Returns
    -------
    :py:class:`xarray.Dataset`
        A subsetted, processed, and reformatted version of `input_store`, with values
        from `constants` having been added.

    Notes
    -----
    * Most often, you will want to use the py:func:`.load_ciam_inputs` function, which
      calls py:func:`.io.prep_sliiders` under the hood.
    * The SLIIDERS-ECON-like dataset may contain cross-sectional variables
      reflecting the geographic and elevational distribution of exposure in a certain
      year (`pop_YYYY` and `K_YYYY`) plus scaling variables to define how these evolve
      over time (`pop_scale` and `K_scale`). This is how  SLIIDERS-ECON 1.0 is
      constructed because it does not consider within-segment migration across elevation slices in its reference socioeconomic trajectories (i.e. in the absence of SLR).
      An alternative specification that this function will handle is one in which `pop`
      and `K` variables within the input dataset explicitly provide estimates of
      exposure for each year, scenario, segment, and elevation. This allows for more flexible assumptions regarding future internal migration but has two drawbacks.
      First, the size of the input dataset is greatly increased. Second, if using pyCIAM
      in "probabilistic" mode, in which you utilize a 2D spline function created by
      :py:func:`.create_surge_lookup`, you must specify a separate lookup table for each scenario and year rather than using just one table.
    """
    inputs_all = xr.open_zarr(input_store, chunks=None).sel(selectors, drop=True)

    inputs = inputs_all.sel({seg_var: seg_vals})
    inputs = _s2d(inputs).assign(constants)

    # assign country level vars to each segment
    for v in inputs.data_vars:
        if "country" in inputs[v].dims:
            inputs[v] = inputs[v].sel(country=inputs.seg_country, drop=True)

    if "vsl" not in inputs.data_vars:
        if "ref_income" in inputs:
            ref_income = inputs.ref_income
        else:
            ref_income = inputs_all.ypcc.sel(country="USA", drop=True)
        ref_income = ref_income.astype("float64").load()
        inputs["vsl"] = (
            inputs.vsl_ypc_mult
            * ref_income
            * (inputs.ypcc / ref_income) ** inputs.vsl_inc_elast
        )

    if "pop" not in inputs.data_vars:
        exp_year = [
            v for v in inputs.data_vars if v.startswith("pop_") and "scale" not in v
        ]
        assert len(exp_year) == 1, exp_year
        exp_year = exp_year[0].split("_")[1]
        pop_var = "pop_" + exp_year
        inputs["pop"] = inputs[pop_var] * inputs.pop_scale
        inputs = inputs.drop(pop_var)
    if "K" not in inputs.data_vars:
        K_var = "K_" + exp_year
        inputs["K"] = inputs[K_var] * inputs.K_scale
        inputs = inputs.drop(K_var)
    if "dfact" not in inputs.data_vars:
        inputs["dfact"] = (1 / (1 + inputs.dr)) ** (inputs.year - inputs.year.min())

    if "landrent" or "ypc" not in inputs.data_vars:
        popdens = (inputs.pop / inputs.landarea).fillna(0)
        if "landrent" not in inputs.data_vars:
            coastland_scale = np.minimum(
                1,
                np.maximum(
                    inputs.min_coastland_scale,
                    np.log(1 + popdens) / np.log(25),
                ),
            )
            inputs["landrent"] = inputs.interior * coastland_scale * inputs.dr

        if "ypc" not in inputs.data_vars:
            ypc_scale = np.maximum(
                inputs.min_ypc_scale,
                (popdens / inputs.ypc_scale_denom) ** inputs.ypc_scale_elast,
            )
            inputs["ypc"] = ypc_scale * inputs.ypcc

    return inputs.drop(
        [
            "pop_scale",
            "K_scale",
            "interior",
            "dr",
            "min_coastland_scale",
            "min_ypc_scale",
            "ypc_scale_denom",
            "ypc_scale_elast",
            "vsl_ypc_mult",
            "vsl_inc_elast",
        ],
        errors="ignore",
    )


def _load_scenario_mc(slr_store, include_ncc=True, include_cc=True):
    scen_mc_filter = (
        xr.open_zarr(slr_store, chunks=None)[["scenario", "mc_sample_id"]]
        .to_dataframe()
        .sort_values(["scenario", "mc_sample_id"])
        .index
    )

    if include_ncc:
        scen_mc_filter = scen_mc_filter.append(
            pd.MultiIndex.from_product(
                (
                    ["ncc"],
                    scen_mc_filter.get_level_values("mc_sample_id")
                    .unique()
                    .sort_values(),
                ),
                names=["scenario", "mc_sample_id"],
            )
        )

    if not include_cc:
        scen_mc_filter = scen_mc_filter[
            scen_mc_filter.get_level_values("scenario") == "ncc"
        ]
    return scen_mc_filter


def _load_lslr_for_ciam(
    slr_store,
    site_id,
    interp_years=None,
    scen_mc_filter=None,
    include_ncc=True,
    include_cc=True,
):

    if scen_mc_filter is None:
        scen_mc_filter = _load_scenario_mc(
            slr_store,
            include_ncc=include_ncc,
            include_cc=include_cc,
        )

    wcc = scen_mc_filter.get_level_values("scenario") != "ncc"
    scen_mc_ncc = scen_mc_filter[~wcc].droplevel("scenario").values
    scen_mc_xr_wcc = (
        scen_mc_filter[wcc]
        .to_frame()
        .reset_index(drop=True)
        .rename_axis(index="scen_mc")
        .to_xarray()
    )

    slr = _s2d(
        xr.open_zarr(slr_store, chunks=None)[["lsl_msl00", "lsl_ncc_msl00"]].sel(
            site_id=site_id, drop=True
        )
    ).drop(["lat", "lon"], errors="ignore")

    # select only the scenarios we wish to model
    if len(scen_mc_xr_wcc.scen_mc):
        slr_out = slr.lsl_msl00.sel(
            scenario=scen_mc_xr_wcc.scenario, mc_sample_id=scen_mc_xr_wcc.mc_sample_id
        ).set_index(scen_mc=["scenario", "mc_sample_id"])
    else:
        slr_out = xr.DataArray([], dims=("scen_mc",), coords={"scen_mc": []})

    if len(scen_mc_ncc):
        slr_ncc = (
            slr.lsl_ncc_msl00.sel(mc_sample_id=scen_mc_ncc)
            .expand_dims(scenario=["ncc"])
            .stack(scen_mc=["scenario", "mc_sample_id"])
        )
        slr_out = xr.concat((slr_out, slr_ncc), dim="scen_mc").sel(
            scen_mc=scen_mc_filter
        )

    # interpolate to yearly
    slr_out = slr_out.reindex(
        year=np.concatenate(([2000], slr.year.values)),
        fill_value=0,
    )

    if interp_years is not None:
        slr_out = slr_out.interp(year=interp_years)
    return slr_out


def create_template_dataarray(dims, coords, chunks, dtype="float32", name=None):
    """A utility function helpful for creatting an empty, dask-backed
    :py:class:`xarray.DataArray` with specific dimensions, coordinates, dtype, name,
    and chunk structure. This is useful for "probabilistic" mode of pyCIAM, in which
    you will run the model on a large ensemble of SLR trajectory realizations. In this
    case, we save an empty Zarr store and then parallelize the model runs across
    regions, with each processor writing to a region within the template store.

    Parameters
    ----------
    dims : list of str
        Dimensions to create.
    coords : dict
        Keys are values in `dims`. Values are a list of values along each dimension.
    chunks : dict
        Keys are values in `dims`. Values are ints defining the chunksize along each
        dimension.
    dtype : str or numpy dtype
        Passed to :py:class:`dask.array.empty`. Defines the dtype of the output array.
    name : str, optional
        Defines the name of the output array.

    Returns
    -------
    :py:class:`xarray.DataArray`
        An empty dask-backed DataArray.
    """
    lens = {k: len(v) for k, v in coords.items()}
    return xr.DataArray(
        da.empty(
            [lens[k] for k in dims], chunks=[chunks[k] for k in dims], dtype=dtype
        ),
        dims=dims,
        coords={k: v for k, v in coords.items() if k in dims},
        name=name,
    )


def create_template_dataset(var_dims, coords, chunks, dtypes):
    """A utility function helpful for creatting an empty, dask-backed
    :py:class:`xarray.Dataset` with specific variables, dimensions, coordinates, dtypes,
    and chunk structure. This is useful for "probabilistic" mode of pyCIAM, in which
    you will run the model on a large ensemble of SLR trajectory realizations. In this
    case, we save an empty Zarr store and then parallelize the model runs across
    regions, with each processor writing to a region within the template store.

    Parameters
    ----------
    var_dims : dict
        Keys are variable names. Values are lists of str, defining the dimensions
        that correspond to each variable.
    coords : dict
        Keys are values in `dims`. Values are a list of values along each dimension.
    chunks : dict
        Keys are values in `dims`. Values are ints defining the chunksize along each
        dimension.
    dtypes : dict of str or numpy dtype
        Defines the dtypes of each variable.

    Returns
    -------
    :py:class:`xarray.Dataset`
        An empty dask-backed Dataset.
    """
    das = {}
    for varname, dims in var_dims.items():
        das[varname] = create_template_dataarray(
            dims,
            {k: v for k, v in coords.items() if k in dims},
            {k: v for k, v in chunks.items() if k in dims},
            dtype=dtypes[varname],
        )
    return xr.Dataset(das)


def check_finished_zarr_workflow(
    finalstore=None,
    tmpstore=None,
    varname=None,
    final_selector={},
    mask=None,
):
    """Check if a workflow that writes to a particular region of a zarr store has
    already run. This is useful when running pyCIAM in "probabilistic" mode across a
    large ensemble of SLR trajectories, using a large, and potentially unstable Dask
    cluster to parallelize computations across regions. If a Dask worker dies, all tasks
    it has completed will be re-computed by another worker. This function is used at the
    start of a mapped computation to check if the work has already been completed and
    saved to disk. If it has, you may wish to skip re-computing it.

    In some cases, we write each worker's individual computations to temporary
    individual zarr stores, and then process different groups of them, writing the
    output of this second stage to regions of a final store. This occurs, for example,
    when the `regions` modeled within CIAM are subsets of decision-making `segment`
    agents. In this case, the first stage of pyCIAM is to calculate costs for all
    adaptation options for each region. In the second stage, we aggregate all regions
    within a segment, sum total costs, and choose an optimal adaptation strategy based
    on the lowest NPV across the segment. We then save the resulting optimal adaptation
    results to a region of a final zarr store.

    This function is designed to handle such a situation, and can check for the
    existence of both the temporary file associated with the first-stage, as well as the
    region of the final zarr store.

    Parameters
    ----------
    finalstore : Path-like, optional
        If specified, path to a final zarr store within which the calling worker is
        going to write to a specific region. If None, do not check for the presence of
        this store.
    tmpstore : Path-like, optional
        If specified, path to a temporary zarr store that will be saved by this task. If
        None, do not check for the presence of this store.
    varname : str
        Name of variable to check within the output zarrs.
    final_selector : dict, optional
        Only necessary if `finalstore` is not None. Keys are dimensions of the Zarr
        store located at `finalstore`. Values are passed to `:py:meth:DataArray.sel`
        before checking for the presence of previously written data.
    mask : DataArray-like of bool
        Mask to use before checking for all non-null values. Mask is applied to both the
        store at `finalstore` and `tmpstore`. `final_selector`, if specified, is
        applied to `finalstore` before `mask`.

    Returns
    -------
    bool :
        True if the `varname` variable within the Zarr store at `finalstore` and/or
        `tmpstore` has all non-null values, after applying `final_selector` and/or
        `mask`. If both `finalstore` and `tmpstore` are specified, both must
        contain null values for this function to return False.
    """
    finished = False
    temp = False
    if finalstore is not None:
        finished = xr.open_zarr(finalstore, chunks=None)[varname].sel(
            final_selector, drop=True
        )
        if mask is not None:
            finished = finished.where(mask, 1)
        finished = finished.notnull().all().item()
    if finished:
        return True
    if tmpstore is not None:
        if tmpstore.fs.isdir(tmpstore.root):
            try:
                temp = xr.open_zarr(tmpstore, chunks=None)
                if mask is not None:
                    temp = temp.where(mask, 1)
                if (
                    varname in temp.data_vars
                    and "year" in temp.dims
                    and len(temp.year) > 0
                ):
                    finished = temp[varname].notnull().all().item()
            except Exception:
                ...
    return finished


def save_to_zarr_region(ds_in, store, already_aligned=False):
    """Wrapper around :py:method:`xarray.Dataset.to_zarr` when specifying the `region`
    kwarg. This function allows you to avoid boilerplate to figure out the integer slice
    objects needed to pass as `region` when calling `:py:meth:xarray.Dataset.to_zarr`.

    Parameters
    ----------
    ds_in : :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
        Dataset or DataArray to save to a specific region of a Zarr store
    store : Path-like
        Path to Zarr store
    already_aligned : bool, default False
        If True, assume that the coordinates of `ds_in` are already ordered the same
        way as those of `store`. May save some computation, but will miss-attribute
        values to coordinates if set to True when coords are not aligned.

    Returns
    -------
    None :
        No return value but `ds_in` is saved to the appropriate region of `store`.

    Raises
    ------
        ValueError
            If `ds_in` is an unnamed DataArray and `store` has more than one variable.
        AssertionError
            If any coordinate values of `ds_in` are not contiguous within `store`.
    """
    ds_out = xr.open_zarr(store, chunks=None)

    # convert dataarray to dataset if needed
    if isinstance(ds_in, xr.DataArray):
        if ds_in.name is not None:
            ds_in = ds_in.to_dataset()
        else:
            if len(ds_out.data_vars) != 1:
                raise ValueError(
                    "``ds_in`` is an unnamed DataArray and ``store`` has more than one "
                    "variable."
                )
            ds_in = ds_in.to_dataset(name=list(ds_out.data_vars)[0])

    # align
    for v in ds_in.data_vars:
        ds_in[v] = ds_in[v].transpose(*ds_out[v].dims).astype(ds_out[v].dtype)

    # find appropriate regions
    alignment_dims = {}
    regions = {}
    for r in ds_in.dims:
        if len(ds_in[r]) == len(ds_out[r]):
            alignment_dims[r] = ds_out[r].values
            continue
        alignment_dims[r] = [v for v in ds_out[r].values if v in ds_in[r].values]
        valid_ixs = np.arange(len(ds_out[r]))[ds_out[r].isin(alignment_dims[r]).values]
        n_valid = len(valid_ixs)
        st = valid_ixs[0]
        end = valid_ixs[-1]
        assert (
            end - st == n_valid - 1
        ), f"Indices are not continuous along dimension {r}"
        regions[r] = slice(st, end + 1)

    # align coords
    if not already_aligned:
        ds_in = ds_in.sel(alignment_dims)

    ds_in.drop_vars(ds_in.coords).to_zarr(store, region=regions)


def load_ciam_inputs(
    input_store,
    slr_store,
    params,
    seg_vals,
    seg_var="seg",
    surge_lookup_store=None,
    ssp=None,
    iam=None,
    scen_mc_filter=None,
    include_ncc=True,
    include_cc=True,
):
    """Load, process, and format all inputs needed to run pyCIAM.

    Parameters
    ----------
    input_store : Path-like
        Path to zarr store that contains the SLIIDERS-ECON dataset (or one formatted
        similarly. May be a :py:class:`fsspec.FSMap` object.
    slr_store : Path-like
        Path to zarr store that contains the SLIIDERS-SLR dataset (or one formatted
        similarly. May be a :py:class:`fsspec.FSMap` object.
    params : dict
        Dictionary of model parameters, typically loaded from a JSON file. See
        :file:`../params.json` for an example of the required parameters.
    seg_vals : list of str
        Defines the subset of regions (along dimension `seg_var`) that the function
        will prep. Subsets are used to run CIAM in parallel.
    seg_var : str, default "seg_var"
        The name of the dimension in `input_store` along which the function will
        subset using `seg_vals`
    surge_lookup_store : Path-like, optional
        If not None, will also load and process data from an ESL impacts lookup table
        (see `lookup.create_surge_lookup`). If included in a call to
        :py:func:`.calc_costs`, including this information will allow
        pyCIAM to run more quickly through the use of a 2D linear spline, rather than
        explicitly estimating ESL-related losses for each segment, year, socioeconomic
        scenario, and SLR trajectory.
    ssp : str or list, optional
        If specified, load only the Shared Socioeconomic Pathways socioeconomic
        scenarios from SLIIDERS-ECON. If using a similarly-formatted socioeconomic
        variable dataset, this will raise a ValueError if specified and if `ssp` is
        not a dimension of the dataset. If None, ignore the `ssp` dimension (i.e. if
        the dimension exists in the dataset, load all SSPs).
    iam : {"IIASA", "OECD", None}, default None
        If specified, load only the version of GDP growth for each loaded SSP
        corresponding to the specified Integrated Assessment Model. If not specified and
        if using SLIIDERS-ECON, load both IAMs. If using a different dataset without the
        `iam` dimension, ignore this dimension. If not using SLIIDERS-ECON, this will
        raise a ValueError if specified and if `iam` is not a dimension of the
        dataset.
    scen_mc_filter : :py:class:`pandas.MultiIndex`, optional
        A list of paired `scenario` (str) and `mc_sample_id` (int) values that
        specify a subset of the individual SLR trajectories contained in the zarr store
        at `sliiders_slr_store`. If None, run all scenario X MC sample combinations.
    include_ncc : bool, default True
        If True, include realizations from the "no-climate-change" SLR scenario (i.e.
        scenarios that include no climate change-driven contributions to SLR).
    include_cc : bool, default True
        If True, include realizations from the various climate change-driven SLR
        trajectories in SLIIDERS-SLR (or a similarly formatted input dataset). Set to
        False when running the pyCIAM "spinup" period, in which initial adaptation is
        estimated through optimizing in the no-climate change scenario.

    Returns
    -------
    inputs : :py:class:`xarray.Dataset`
        A processed and formatted dataset of socioeconomic variables used in pyCIAM.
    slr : :py:class:`xarray.Dataset`
        A processed and formatted dataset of sea level rise variables used in pyCIAM.
    surge : :py:class:`xarray.Dataset`
        A processed and formatted dataset of extreme sea level damages used in pyCIAM.
        If `surge_lookup_store` is None, this output will be None

    Raises
    ------
    ValueError
        If `ssp` or `iam` is specified and the corresponding variables are not
        present in the Zarr store located at `input_store`.
    """
    selectors = {}
    if ssp is not None:
        selectors["ssp"] = ssp
    if iam is not None:
        selectors["iam"] = iam
    inputs = prep_sliiders(
        input_store,
        seg_vals,
        constants=params,
        seg_var=seg_var,
        selectors=selectors,
    )

    if seg_var != "seg":
        inputs = inputs.drop("seg", errors="ignore").rename({seg_var: "seg"})
    inputs.load()

    # get surge lookup table
    if surge_lookup_store is not None:
        surge = (
            xr.open_zarr(surge_lookup_store, chunks=None)
            .sel({seg_var: seg_vals})
            .rename({seg_var: "seg"})
            .load()
        )
    else:
        surge = None

    # get SLR
    site_ids = inputs.SLR_site_id.values
    slr = (
        _load_lslr_for_ciam(
            slr_store,
            np.unique(site_ids),
            interp_years=inputs.year.values,
            scen_mc_filter=scen_mc_filter,
            include_ncc=include_ncc,
            include_cc=include_cc,
        )
        .sel(site_id=site_ids)
        .rename(site_id="seg")
    )
    slr["seg"] = seg_vals

    return inputs, slr, surge


def load_diaz_inputs(input_store, seg_vals, params, include_ncc=True, include_cc=True):
    """Load the original inputs used in Diaz 2016.

    Parameters
    ----------
    input_store : Path-like
        Path to zarr store that contains the formatted
        `Diaz 2016 inputs <https://github.com/delavane/CIAM/blob/master/CIAMdata.gdx>`_.
        May be a :py:class:`fsspec.FSMap` object.
    seg_vals : list of str
        Defines the subset of regions (along dimension `seg_var`) that the function
        will prep. Subsets are used to run CIAM in parallel.
    params : dict
        Dictionary of model parameters, typically loaded from a JSON file. See
        :file:`../params.json` in the pyCIAM github repository for an example of the
        required parameters.
    include_ncc : bool, default True
        If True, include realizations from the "no-climate-change" SLR scenario (i.e.
        scenarios that include no climate change-driven contributions to SLR).
    include_cc : bool, default True
        If True, include realizations from the various climate change-driven SLR
        trajectories in SLIIDERS-SLR (or a similarly formatted input dataset). Set to
        False when running the pyCIAM "spinup" period, in which initial adaptation is
        estimated through optimizing in the no-climate change scenario.

    Returns
    -------
    inputs : :py:class:`xarray.Dataset`
        A processed and formatted dataset of socioeconomic variables used in pyCIAM.
    slr : :py:class:`xarray.Dataset`
        A processed and formatted dataset of sea level rise variables used in pyCIAM.

    Notes
    -----
    * `surge` is not returned when running the original Diaz 2016 specification, as
      ESL-related losses are estimated via a pre-calculated exponential function
    """
    inputs = prep_sliiders(input_store, seg_vals, constants=params, seg_var="seg")
    ncc_inputs = inputs.rcp_pt.str.startswith("rcp0")
    lsl_ncc = inputs.lsl.isel(rcp_pt=ncc_inputs)
    lsl_wcc = inputs.lsl.isel(rcp_pt=~ncc_inputs)
    slr = xr.concat(
        [i for i, j in ((lsl_ncc, include_ncc), (lsl_wcc, include_cc)) if j],
        dim="rcp_pt",
    )
    inputs = inputs.drop_dims("rcp_pt")
    return inputs, slr



"""This module contains various constants which reference lists of string values
defining cost types and potential adaptation options.

Constants
    RLIST : list of str
        List of the discrete retreat options, corresponding to different return heights
        of extreme sea level. Other than ``1``, the suffixes in this list must be included
        in the `surge_heights` coordinates of SLIIDERS-ECON (or a similarly formatted
        dataset).
    PLIST : list of str
        Same as RLIST except for protection. pyCIAM does not allow for protection to the
        maximum projected local MSL height during an adaptation period, as seen with
        `retreat1` in `RLIST`.
    SOLVCASES : list of str
        A list of all adaptation options available to decision-making coastal segment
        agents, including the "reactive retreat" option (`noAdaptation`).
    CASES : list of str
        Same as SOLVCASES, but including the `optimalfixed` case, which is equivalent
        to one of the other `CASES`. It represents the optimal adaptation choice,
        which will be segment-specific.
    CASE_DICT : dict
        Keys are values of `SOLVCASES`. Values are integer index values.
    COSTTYPES : list of str
        A list of all cost types calculated in pyCIAM.
"""

# Retreat cases
RLIST = ["retreat1", "retreat10", "retreat100", "retreat1000", "retreat10000"]

# Protect cases
PLIST = ["protect10", "protect100", "protect1000", "protect10000"]

# All adaptation cases to calculate
SOLVCASES = ["noAdaptation"] + PLIST + RLIST  # cases to calc
CASES = SOLVCASES + ["optimalfixed"]  # all cases + slot for optimal (least cost) case

# data dictionary for different adaptation cases
CASE_DICT = {k: kx for kx, k in enumerate(SOLVCASES)}

# Cost types for which damages are calculated in pyCIAM()
COSTTYPES = [
    "wetland",
    "inundation",
    "relocation",
    "protection",
    "stormCapital",
    "stormPopulation",
]



"""This module is used to define the fractional damage functions used in pyCIAM that
relate storm surge depth to mortality and physical capital loss. As of April 7, 2022,
the only functions available are those included in Diaz 2016. These define the
"resilience-unadjusted" fractional damages. That is, for a region with resiliendce
factor (:math:`\rho`) of 1, what fraction of the exposed population and/or physical capital
will be lost conditional on a storm surge depth.

At the moment, functions must have an analytical integral and are actually defined by
their integral, as can be seen with the suffix ``_i``.

To add a new damage function, simply define it by its integral where depth is in units
of meters. It must contain two mandatory arguments which refer to the bounds of the
definite integral. Kwargs relevant for this particular damage function are optional.
The output should be "fractional loss" assuming homogeonous distribution of
capital/population within a grid cell defined by the two bounds.

Example
-------
Let's say you wanted a linear damage function in depth where losses increased by 10% per
meter. The integral is this constant, so the damage function would defined simply as:

.. code-block:: python

    def linear_damage_func_i(depth_st, depth_end):
        return 0.1

This function would of course result in losses greater than 100% for depths over 10m and
is just shown here as an example
"""

import numpy as np


def diaz_ddf_i(depth_st, depth_end):
    """Integral of depth / (1 + depth), as used in Diaz 2016, assuming unit resilience
    (:math:`\rho`).

    Parameters
    ----------
    depth_{st,end} : float
        Define the lower and upper bounds of a definite integral

    Returns
    -------
    float :
        Fractional losses of physical capital, assuming homogenous distribution between
        `depth_st` and `depth_end`.
    """
    return depth_end - depth_st + np.log(depth_st + 1) - np.log(depth_end + 1)


def diaz_dmf_i(depth_st, depth_end, floodmortality=None):
    """Integral of mortality damage function as used in Diaz 2016, assuming unit
    resilience (:math:`\rho`). It is just a constant fraction conditional on a unit of exposure being inundated. Note that kwargs are not optional and will raise an error
    if not specified when called.

    Parameters
    ----------
    depth_{st,end} : float
        Define the lower and upper bounds of a definite integral
    floodmortality : float
        Fractional damage conditional on a unit of exposure being inundated.

    Returns
    -------
    float :
        Fractional mortality, assuming homogenous distribution between `depth_st` and `depth_end`
    """
    return floodmortality * (depth_end - depth_st)



"""This private module contains miscellaneous functions to support pyCIAM"""

from pathlib import Path

import fsspec
import numpy as np
import pandas as pd
import xarray as xr


def _str_to_mapper(path):
    if isinstance(path, (Path, str)):
        fs = fsspec.filesystem("file")
        return fs.get_mapper(path)
    return path


def _s2d(ds):
    for v in ds.data_vars:
        if ds[v].dtype == "float32":
            ds[v] = ds[v].astype("float64")
    return ds


def _get_planning_period_map(years, at_start):
    return (
        (years < xr.DataArray(at_start[1:], dims=["at"], coords={"at": at_start[:-1]}))
        .reindex(at=at_start, fill_value=True)
        .idxmax("at")
    )


def _pos(x):
    return np.maximum(x, 0)


def _get_lslr_plan_data(
    lslr,
    surge_heights,
    planning_periods,
    diaz_protect_height=False,
    diaz_lslr_plan=False,
    diaz_negative_retreat=False,
    min_R_noadapt=None,
):

    lslr_plan_noadapt = lslr.copy()
    if min_R_noadapt is not None:
        lslr_plan_noadapt = np.maximum(lslr_plan_noadapt, min_R_noadapt)

    # prohibit negative adaptation
    if diaz_negative_retreat:
        lslr_plan_noadapt = _pos(lslr_plan_noadapt)
    else:
        for i in range(1, len(lslr_plan_noadapt.year)):
            lslr_plan_noadapt[{"year": i}] = lslr_plan_noadapt.isel(
                year=slice(None, i + 1)
            ).max("year")

    # Planning slr height for retreat/protect scenarios. Diaz 2016 uses lslr at the
    # start of the next planning period as the design height for this planning period.
    # Updated model uses maximum of the LSLR within this planning period
    if diaz_lslr_plan:
        plan_years = np.unique(planning_periods)
        design_years = np.concatenate(
            (plan_years[1:], lslr.year.isel(year=[-1]).values)
        )
        lslr_plan = lslr.sel(year=design_years).rename(year="at")
        lslr_plan["at"] = plan_years
    else:
        lslr_plan = lslr.groupby(planning_periods).max().rename("lslr_plan")

    # hack to reduce surge height by 50% for protect 10 as in Diaz2016
    if diaz_protect_height:
        surge_heights_p = surge_heights.where(
            surge_heights.return_period != 10, surge_heights / 2
        )
    else:
        surge_heights_p = surge_heights

    surge_heights = xr.concat(
        (surge_heights, surge_heights_p),
        dim=pd.Index(["retreat", "protect"], name="adapttype"),
    )

    # calculate retreat and protect heights
    RH_heights = lslr_plan + surge_heights

    # add in the MSL-level advanced planning return height
    RH_heights = xr.concat(
        (lslr_plan.expand_dims(return_period=[1]), RH_heights), dim="return_period"
    )

    # Prohibit negative adaptation
    if diaz_negative_retreat:
        RH_heights = _pos(RH_heights)
    else:
        for i in range(1, len(RH_heights.at)):
            RH_heights[{"at": i}] = RH_heights.isel(at=slice(None, i + 1)).max("at")

    # set initial RH_heights to 0 (e.g.
    # assuming no retreat or protection anywhere such that both w/ and w/o climate
    # change scenarios are charged for this adaptation). Alternative would be to set
    # protections up to current surge height already exist (e.g. assuming perfect
    # adaptation already).
    RH_heights_prev = RH_heights.shift(at=1, fill_value=0)

    return (
        lslr_plan_noadapt,
        RH_heights,
        RH_heights_prev,
    )



"""This module contains functions related to creating a storm surge lookup table used
when running pyCIAM in "probabilistic" mode (i.e. running on many thousands of Monte
Carlo samples of sea level rise trajectories). In this mode, calculating storm surge
damages for each elevation slice, each year, each segment, each socioeconomic
trajectory, and each SLR sample is too computationally intensive. This is due to the
need to numerically integrate in each of these cases. Instead, we calculate fractional
losses for a number of storm surge heights between the minimum and maximum that will
occur across the entire SLR ensemble and use linear interpolation to create a linear
spline function that allows for rapid estimation of storm surge losses across a large
ensemble of SLR simulations. As currently configured and used, this lookup table relies
on the assumption of a fixed elevational distribution of capital and population that is
scaled homogeonously over time. However, if one wanted to incorporate within-country
migration into future socioeconomic trajectories, one could do so through the use of
year-specific lookup tables (a.k.a. spline functions).

Public Functions:
    create_surge_lookup
"""


import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from .._utils import _get_lslr_plan_data, _get_planning_period_map, _s2d, _str_to_mapper
from ..io import _load_lslr_for_ciam, save_to_zarr_region
from ._calc import _calc_storm_damages_no_resilience, _get_surge_heights_probs
from .damage_funcs import diaz_ddf_i, diaz_dmf_i


def _get_exposure_vars(ds):
    return [
        d
        for d in ds.data_vars
        if (d.startswith("K_") or d.startswith("pop_")) and not d.endswith("_scale")
    ]


def _get_lslr_rhdiff_range(
    sliiders_econ_store,
    sliiders_slr_store,
    seg_var,
    seg_vals,
    at_start,
    n_interp_pts_lslr,
    n_interp_pts_rhdiff,
    quantiles=None,
    interp_years=None,
    scen_mc_filter=None,
    include_cc=True,
    include_ncc=True,
):
    """Get the range of lslr and rhdiff that we need to model to cover the full range
    across scenario/mcs. The minimum LSLR value we'll need to model for the purposes of
    assessing storm damage is the minimum across sites of: the site-level maximum of "0
    minus the s10000 surge height" and "the minimum projected LSLR for all of the
    scenario/mcs we use in our binned LSL dataset". The maximum LSLR value is the
    maximum experienced at any site in any year for all of the sceanrio/mcs we use in
    the binned LSL dataset."""

    pc_in = _s2d(
        xr.open_zarr(sliiders_econ_store, chunks=None).sel({seg_var: seg_vals})
    )

    if interp_years is None:
        interp_years = pc_in.year.values

    # list out which sites are actually used for CIAM segments
    site_ids = pc_in.SLR_site_id
    unique_site_ids = np.unique(site_ids)

    pc_in.surge_height.load()
    pc_in.gumbel_params.load()

    # get max surge height for this seg-adm
    smax = pc_in.surge_height.isel(return_period=-1)

    lslr = _load_lslr_for_ciam(
        sliiders_slr_store,
        unique_site_ids,
        interp_years=interp_years,
        scen_mc_filter=scen_mc_filter,
        include_cc=include_cc,
        include_ncc=include_ncc,
    ).sel(site_id=site_ids, drop=True)

    if quantiles is not None:
        lslr = lslr.unstack().quantile(quantiles, dim="mc_sample_id")

    # get the max LSLR experienced
    max_lslr = lslr.max([d for d in lslr.dims if d != seg_var]).sel({seg_var: seg_vals})

    # find the min LSLR we need to model. This is the max of "min lslr" and the lowest
    # elevation in this seg-adm minus the max surge height
    exp_vars = _get_exposure_vars(pc_in)
    assert len(exp_vars) == 2
    min_lslr = pc_in[exp_vars].to_array("tmp").sum("tmp") > 0
    if min_lslr.sum() == 0:
        return None
    min_lslr = pc_in.elev_bounds.sel(bound="lower", drop=True).where(min_lslr).min()
    min_lslr = np.maximum(
        lslr.min([d for d in lslr.dims if d != seg_var]), min_lslr - smax
    )

    # ensure that max > min to enable interpolation even when no damage is possible
    max_lslr = max_lslr.where(max_lslr > min_lslr, min_lslr + 0.01 * n_interp_pts_lslr)

    at = _get_planning_period_map(lslr.year, at_start)

    (
        _,
        RH_heights,
        _,
    ) = _get_lslr_plan_data(lslr, pc_in.surge_height, at)

    # allow for some floating point error in rh_diff_max
    rh_diff_max = (
        RH_heights.sel(at=at).isel(return_period=-2, drop=True).max("adapttype") - lslr
    ).max([d for d in lslr.dims if d != seg_var]) + 2 * np.finfo("float32").eps

    # occasionally, the gumbel fit was negative, so we set the 1-year return to 0
    assert (rh_diff_max > 0).all()

    return xr.Dataset(
        {
            "lslr_by_seg": (
                ("lslr", seg_var),
                np.linspace(min_lslr, max_lslr, n_interp_pts_lslr),
            ),
            "rh_diff_by_seg": (
                ("rh_diff", seg_var),
                np.linspace(0, rh_diff_max, n_interp_pts_rhdiff),
            ),
        },
        coords={
            "lslr": np.arange(n_interp_pts_lslr),
            "rh_diff": np.arange(n_interp_pts_lslr),
            seg_var: pc_in[seg_var].values,
        },
    )


def _create_surge_lookup_skeleton_store(
    sliiders_econ_store,
    n_interp_pts_lslr,
    n_interp_pts_rhdiff,
    surge_lookup_store,
    seg_chunksize=1,
    seg_var="seg",
    force_overwrite=True,
):
    pc_in = xr.open_zarr(sliiders_econ_store)

    to_save = xr.DataArray(
        da.empty(
            (len(pc_in[seg_var]), n_interp_pts_lslr, n_interp_pts_rhdiff, 2, 2),
            chunks=(seg_chunksize, -1, -1, -1, -1),
        ),
        dims=[seg_var, "lslr", "rh_diff", "costtype", "adapttype"],
        coords={
            seg_var: pc_in[seg_var].values,
            "lslr": np.arange(n_interp_pts_lslr),
            "rh_diff": np.arange(n_interp_pts_rhdiff),
            "adapttype": ["retreat", "protect"],
            "costtype": ["stormCapital", "stormPopulation"],
        },
    ).to_dataset(name="frac_losses")
    to_save["rh_diff_by_seg"] = (
        (seg_var, "rh_diff"),
        da.empty(
            (len(to_save[seg_var]), len(to_save.rh_diff)),
            chunks=(to_save.chunks[seg_var], to_save.chunks["rh_diff"]),
        ),
    )
    to_save["lslr_by_seg"] = (
        (seg_var, "lslr"),
        da.empty(
            (len(to_save[seg_var]), len(to_save.rh_diff)),
            chunks=(to_save.chunks[seg_var], to_save.chunks["rh_diff"]),
        ),
    )
    if force_overwrite:
        to_save.to_zarr(surge_lookup_store, compute=False, mode="w")
    elif not surge_lookup_store.fs.isdir(surge_lookup_store.root):
        to_save.to_zarr(surge_lookup_store, compute=False)
    return to_save


def _save_storm_dam(
    seg_vals,
    seg_var="seg",
    sliiders_econ_store=None,
    sliiders_slr_store=None,
    surge_lookup_store=None,
    at_start=np.arange(2000, 2100, 10),
    n_interp_pts_lslr=100,
    n_interp_pts_rhdiff=100,
    ddf_i=diaz_ddf_i,
    dmf_i=diaz_dmf_i,
    ddf_kwargs={},
    dmf_kwargs={},
    quantiles=None,
    scen_mc_filter=None,
):
    """Function to map over each chunk to run through damage calcs."""
    diff_ranges = _get_lslr_rhdiff_range(
        sliiders_econ_store,
        sliiders_slr_store,
        seg_var,
        seg_vals,
        at_start,
        n_interp_pts_lslr,
        n_interp_pts_rhdiff,
        quantiles=quantiles,
        scen_mc_filter=scen_mc_filter,
    )

    if diff_ranges is None:
        template = xr.open_zarr(surge_lookup_store, chunks=None).sel(
            {seg_var: seg_vals}
        )
        template = xr.zeros_like(template)

        # these must be unique otherwise interp function will raise error
        template["lslr_by_seg"] = (
            (seg_var, "lslr"),
            np.tile(np.arange(len(template.lslr))[np.newaxis, :], (len(seg_vals), 1)),
        )
        template["rh_diff_by_seg"] = (
            (seg_var, "rh_diff"),
            np.tile(
                np.arange(len(template.rh_diff))[np.newaxis, :], (len(seg_vals), 1)
            ),
        )
        if surge_lookup_store is None:
            return template
        save_to_zarr_region(template, surge_lookup_store)
        return None

    pc_in = xr.open_zarr(sliiders_econ_store, chunks=None)
    exp_vars = _get_exposure_vars(pc_in)
    pc_in = (
        pc_in[
            [
                "surge_height",
                "gumbel_params",
                "elev_bounds",
            ]
            + exp_vars
        ]
        .reset_coords(drop=True)
        .sel({seg_var: seg_vals}, drop=True)
        .load()
        .astype("float64")
    )
    pc_in = pc_in.rename({k: k.split("_")[0] for k in exp_vars})

    exp0 = pc_in[["K", "pop"]]
    exp0_tot = exp0.sum("elev")
    exp_frac = (exp0 / exp0_tot).where(exp0_tot != 0, 0)
    pc_in = pc_in.drop(
        [
            "K",
            "pop",
        ]
    )
    del exp0, exp0_tot

    this_max_surge_height = pc_in.surge_height.isel(return_period=-1, drop=True)

    surge_heights_to_model = _get_surge_heights_probs(
        0,
        this_max_surge_height,
        pc_in.gumbel_params,
    )

    res = []
    for R, H in [
        (diff_ranges.lslr_by_seg + diff_ranges.rh_diff_by_seg, 0),
        (0, diff_ranges.lslr_by_seg + diff_ranges.rh_diff_by_seg),
    ]:
        res.append(
            _calc_storm_damages_no_resilience(
                diff_ranges.lslr_by_seg + surge_heights_to_model.surge,
                exp_frac,
                pc_in.elev_bounds,
                R,
                H,
                ddf_i,
                dmf_i,
                ddf_kwargs=ddf_kwargs,
                dmf_kwargs=dmf_kwargs,
                surge_probs=surge_heights_to_model.p,
            ).to_array("costtype")
        )
    res = (
        xr.concat(res, dim=pd.Index(["retreat", "protect"], name="adapttype"))
        .reindex(costtype=["stormCapital", "stormPopulation"])
        .to_dataset(name="frac_losses")
    )

    res = xr.merge((res, diff_ranges))

    if surge_lookup_store is None:
        return res

    # identify which index to save to in template zarr
    save_to_zarr_region(res, surge_lookup_store)


def create_surge_lookup(
    sliiders_econ_store,
    sliiders_slr_store,
    surge_lookup_store,
    seg_var,
    at_start,
    n_interp_pts_lslr,
    n_interp_pts_rhdiff,
    ddf_i,
    dmf_i,
    ddf_kwargs={},
    dmf_kwargs={},
    seg_chunksize=1,
    scen_mc_filter=None,
    quantiles=None,
    force_overwrite=False,
    client=None,
    client_kwargs={},
):
    """Create a storm surge lookup table which is used to define a linear spline
    function for each region modeled in pyCIAM. This output is not strictly necessary to
    run pyCIAM but substantially reduces computational expense when running pyCIAM on a
    large probabilistic ensemble of SLR trajectories.

    Parameters
    ----------
    sliiders_{econ,slr}_store : Path-like
        Path to zarr store that contains the SLIIDERS-ECON and SLIIDERS-SLR datasets (or
        ones formatted similarly. May be a :py:class:`fsspec.FSMap` object.
    surge_lookup_store : Path-like
        Path to the output zarr store that will contain the lookup table used to build
        the 2D linear spline function. May be a :py:class:`fsspec.FSMap` object.
    seg_var : str
        The name of the dimension in the SLIIDERS-ECON zarr store specifying each
        *region* to be modeled by pyCIAM. Each region must be nested within each coastal
        segment that serves as an independent decision-making agent, and regions may be
        equivalent to segments. The reason you may wish to have nested regions is to
        be able to aggregate impacts to a different regions than those that are defined
        by the segments.
    at_start : list of int
        A list specifying the starting years of each adpatation period. In pyCIAM, each
        segment chooses a new retreat or protection height at the start of each of these
        periods based on the maximum projected extreme sea level heights at various
        return intervals over the period.
    n_interp_pts_{lslr,rhdiff} : int
        The number of interpolation points to estimate within the output lookup table.
        The lookup table enables 2-dimensional linear interpolation where one dimension
        is defined by local sea level height (lslr) and the other by the difference
        between an assumed extreme sea level height and the height of either retreat or
        protection (rhdiff). This function calculates the maximum and minimum of these
        values across all years and SLR trajectories included in SLIIDERS-SLR and
        generates equally spaced interpolation points.
    {ddf,dmf}_i : function
        Functions defining integrals of the depth-damage and depth-mortality functions
        used to calculate losses from extreme sea levels.
    seg_chunksize : int, default 1
        How many regions to process at once. Larger numbers improve efficiency through
        better vectorization but increase memory footprint.
    scen_mc_filter : :py:class:`pandas.MultiIndex`, optional
        A list of paired `scenario` (str) and `mc_sample_id` (int) values that
        specify a subset of the individual SLR trajectories contained in the zarr store
        at `sliiders_slr_store`. If None, run all scenario X MC sample combinations.
    quantiles : array_like, optional
        If not None, run only specified quantiles of SLR per year, per location, and per
        scenario, across the `mc_sample_id` dimension of the zarr store located at
        `sliiders_slr_store`. If None, run all individual Monte Carlo samples.
    force_overwrite : bool, default False
        Whether to raise an error (True) or overwrite if `surge_lookup_store` already
        contains an existing zarr store.
    client : :py:class:`distributed.Client`, optional
        A dask.distributed Client object that will be used to parallelize the table
        generation process. If not specifiec, outputs will be generated using a single
        processor in series.

    Returns
    -------
    Returns None, but saves storm surge lookup table to `surge_lookup_store`.
    """

    sliiders_econ_store, sliiders_slr_store, surge_lookup_store = list(
        map(
            _str_to_mapper,
            [sliiders_econ_store, sliiders_slr_store, surge_lookup_store],
        )
    )

    to_save = _create_surge_lookup_skeleton_store(
        sliiders_econ_store,
        n_interp_pts_lslr,
        n_interp_pts_rhdiff,
        surge_lookup_store,
        seg_chunksize=seg_chunksize,
        seg_var=seg_var,
        force_overwrite=force_overwrite,
    )
    all_segs = to_save[seg_var].values
    seg_grps = [
        all_segs[i : i + seg_chunksize] for i in range(0, len(all_segs), seg_chunksize)
    ]

    if client is None:
        mapper = map
    else:
        mapper = client.map
    return list(
        mapper(
            _save_storm_dam,
            seg_grps,
            seg_var=seg_var,
            sliiders_econ_store=sliiders_econ_store,
            sliiders_slr_store=sliiders_slr_store,
            surge_lookup_store=surge_lookup_store,
            at_start=at_start,
            n_interp_pts_lslr=n_interp_pts_lslr,
            n_interp_pts_rhdiff=n_interp_pts_rhdiff,
            ddf_i=ddf_i,
            dmf_i=dmf_i,
            quantiles=quantiles,
            scen_mc_filter=scen_mc_filter,
            ddf_kwargs=ddf_kwargs,
            dmf_kwargs=dmf_kwargs,
            **client_kwargs,
        )
    )



"""This module contains the central engine of pyCIAM, in which costs for all adaptation
options are calculated.

Functions
    calc_costs
    select_optimal_case
"""

import numpy as np
import pandas as pd
import xarray as xr

from ._utils import _get_lslr_plan_data, _get_planning_period_map, _pos
from .constants import CASE_DICT, PLIST, RLIST
from .surge._calc import _calc_storm_damages_no_resilience, _get_surge_heights_probs
from .surge.damage_funcs import diaz_ddf_i, diaz_dmf_i


def calc_costs(
    inputs,
    lslr,
    surge_lookup=None,
    elev_chunksize=1,
    ddf_i=diaz_ddf_i,
    dmf_i=diaz_dmf_i,
    ddf_kwargs={},
    dmf_kwargs={},
    diaz_protect_height=False,
    diaz_construction_freq=False,
    diaz_lslr_plan=False,
    diaz_negative_retreat=False,
    diaz_forward_diff=False,
    diaz_zero_costs_in_first_year=False,
    diaz_storm_calcs=False,
    diaz_fixed_vars_for_onetime_cost=False,
    diaz_calc_noadapt_damage_w_lslr=False,
    min_R_noadapt=None,
    return_year0_hts=False,
    return_RH_heights=False,
):
    """This is the central engine of pyCIAM. It calculates costs for all adaptation
    options (a.k.a. *cases*). It does not yet calculate the optimal strategy because
    users may have specified `regions` that are more finely resolved than coastal
    *segments*, which are the decision-making agents in pyCIAM. In this case, this
    function may be parallelized over *regions*, and the results must be aggregated to
    the *segment* level before finding the lowest-cost trajectory.

    Parameters
    ----------
    inputs : :py:class:`xarray.Dataset`
        A processed and formatted version of SLIIDERS-ECON (or similarly formatted
        dataset), ready to be ingested into pyCIAM. See :py:func:`.load_ciam_inputs`
    lslr : :py:class:`xarray.Dataset`
        A processed and formatted version of SLIIDERS-SLR (or similarly formatted
        dataset), ready to be ingested into pyCIAM. See :py:func:`.load_ciam_inputs`
    surge_lookup : :py:class:`xarray.DataArray`, optional
        If not None, a DataArray containing at least ``lslr`` and ``rhdiff`` dimensions.
        This will be used to build a 2D spline function to interpolate extreme sea
        level-related damages, rather than calculating explicitly. This drastically
        improves execution time.
    elev_chunksize : int, default 1
        Number of elevation slices to process simultaneously. Higher numbers improve
        efficiency through vectorization but result in a larger memory footprint.
    ddf_i, dmf_i : func, default :py:func:`.damage_funcs.ddf_i`, :py:func:`.damage_funcs.dmf_i`
        Damage functions relating physical capital loss and monetized mortality arising
        from a certain depth of inundation.
    dmf_kwargs, ddf_kwargs : dict, optional
        Kwargs passed directly to `ddf_i` and `dmf_i`
    diaz_protect_height : bool, default False
        If True, reduce the 1-in-10-year extreme sea level by 50% as in Diaz 2016. This
        hack should not be necessary when using the ESL heights from CoDEC (as in
        SLIIDERS-ECON).
    diaz_construction_freq : bool, default False
        If True, set the lifetime over which the "linear" component of protection
        construction costs (i.e. the component that is proportional to length and not to
        height) is amortized to the length of the adaptation periods. This means that
        the total costs from this linear portion was directly proportional to the number
        of planning periods. Because the default planning periods for pyCIAM are
        shorter, we assume that the lifetime of the seawall foundation that depends on
        this linear part is 50 years (roughly the length of the planning periods in Diaz 2016). This allows  total seawall construction costs to be roughly independent
        of the number of planning periods.
    diaz_lslr_plan : bool, default False
        If True, use the local sea level height at the start of the next planning period
        to define the design height for retreat or protection actions (as in Diaz 2016).
        If False, use the maximum slr height within the current adaptation planning
        period, which may differ if local SLR is not monotonically increasing through
        the adaptation period.
    diaz_negative_retreat : bool, default False
        If True, only prohibit retreat below present-day MSL but allow migration and/or
        protection to lower SLR levels as long as you don't migrate below. This creates
        inconsistencies in the calculation of retreat, protection, and inundation costs
        for regions that experience negative local SLR. If False, prohibit downward
        migration and/or lowering protection heights to preserve consistency of cost
        calculations.
    diaz_forward_diff : bool, default False
        If True, use a mixture of forward and backward difference techniques to
        calculate rates of change for different cost calculations, as in Diaz 2016. If
        False, use backward difference in all contexts.
    diaz_zero_costs_in_first_year : bool, default False
        If True, ignore all costs except wetland loss costs in the first year for all
        cases except "reactive retreat", as in Diaz 2016. If False, account for these
        costs.
    diaz_storm_calcs : bool, default False
        If True, use the pre-calculated exponential functions, parameterized by segment,
        from Diaz 2016 for ESL-related costs. Note that these are as implemented in the
        `associated code <https://github.com/ClimateImpactLab/pyCIAM>`_, which is
        slightly different than described in the paper and assumes constant marginal
        physical capital losses regardless of starting sea level height. If False, use
        either a pre-calculated 2D spline function (if `surge_lookup` is not None) or
        directly calculate ESL costs for each segment, year, elevation, socioeconomic
        scenario, and SLR trajectory. Note that this may only be True if the parameters
        of the exponential function are specified in `inputs`.
    diaz_fixed_vars_for_onetime_cost : bool, default False
        If True, use the price of construction at the start of an adaptation period to
        calculate amortized one-time costs (as in Diaz 2016). If False, account for
        growth in GDP and correlated variables (e.g. land value) over the adaptation
        period.
    diaz_calc_noadapt_damage_w_lslr : bool, default False
        In Diaz 2016, a "spinup" model is used first to calculate starting retreat
        heights assumed for the "reactive retreat" case (a.k.a. "noAdaptation").
        However, when calculating inundation and relocation costs for that reactive
        retreat case, this retreat height is ignored and segments are assumed to only
        retreat to the level of local sea level in a given year (or to the starting sea
        level if future LSLR values become negative). This results in inconsistent
        retreat heights used for inundation and relocation costs than for wetland loss
        and ESL-related losses. This is the approach used if True. If False, stay
        consistent with the retreat height for the reactive retreat scenario case by
        assuming the spinup retreat height for all cost calculations.
    min_R_noadapt : :py:class:`xarray.DataArray`, optional
        If specified, this should be a 1D DataArray indexed by region, which is a
        dimension name that should match that of `inputs`, `lslr`, and/or
        `surge_lookup`. This is the result of having previously run pyCIAM in "spinup"
        mode to determine the initial retreat height for the reactive retreat scenario.
    return_year0_heights : bool, default False
        If True, in addition to costs, output the corresponding retreat or protect
        elevation for each adaptation case in the initial year. This us used when run in
        "spinup" mode. Once the optimal case is chosen for each segment, the
        corresponding height is used as `min_R_noadapt` in a subsequent run of pyCIAM.
    return_RH_heights : bool, default False
        Not typically used, this flag means that results will also include the retreat
        and protection heights for all years. This is useful for the revealed preference
        identification of the non-market costs of relocation performed in Depsky et al.
        2022.

    Returns
    -------
    :py:class:`xarray.Dataset`
        Dataset of costs for each region included in `inputs`, for all scenarios and
        Monte Carlo SLR samples in `lslr`, for all cost types, for all adaptation cases
        *other* than the optimal case, which must be calculated separately. This is to
        allow for arbitrary aggregation of costs from multiple regions within a
        decision-making segment  before calculating the least-cost adaptation pathway.
    :py:class:`xarray.DataArray`
        If `return_year0_heights` is True, also return the first adaptation heights for
        each adaptation case, SLR scenario, and Monte Carlo sample. Used when running
        pyCIAM in "spinup" mode.
    :py:class:`xarray.Dataset`
        If `return_RH_heights` is True, also return the retreat and protection heights
        from all years.

    Notes
    -----
    * Year-specific `surge_lookup` arrays have not yet been implemented, but would be
      necessary if projecting within-segment heterogeneous population, GDPpc, or
      physical capital growth (i.e. differing over elevation slices). This does not
      exist in SLIIDERS-ECON, for which growth is homogeonous within each country.
    """
    # You should have already made sure your inputs and slr are in the same years
    assert (lslr.year == inputs.year).all()

    # if elev_chunksize is None, don't chunk over elev dimension at all
    if elev_chunksize is None:
        elev_chunksize = len(inputs.elev)

    # calculate lower bound of elevation bins
    if "elev_bounds" in inputs.data_vars:
        lb_elev = inputs.elev_bounds.sel(bound="lower", drop=True).copy()
        bin_width = inputs.elev_bounds.sel(bound="upper", drop=True) - lb_elev
    else:
        bin_width = inputs.elev.diff("elev").reindex(elev=inputs.elev).bfill("elev")
        lb_elev = (inputs.elev - bin_width / 2).copy()

    # correct potential numerical rounding issue
    lb_elev[0] = 0

    # get planning period info
    at = _get_planning_period_map(lslr.year, inputs.at_start.values)
    tstept = (
        inputs.year.diff("year", label="lower").reindex(year=inputs.year).ffill("year")
    )
    tstep_at = tstept.groupby(at).sum()

    # Calculate annual local SLR rate btw t periods. This is one calculation that diaz
    # does with a forward difference
    if diaz_forward_diff:
        localrate = _pos(
            (lslr.diff("year", label="lower").reindex(year=lslr.year) / tstept).ffill(
                "year"
            )
        )
    else:
        localrate = _pos(
            (lslr.diff("year").reindex(year=lslr.year) / tstept).bfill("year"),
        )

    # get various planning height values
    (lslr_plan_noadapt, RH_heights, RH_heights_prev,) = _get_lslr_plan_data(
        lslr,
        inputs.surge_height,
        at,
        diaz_protect_height=diaz_protect_height,
        diaz_lslr_plan=diaz_lslr_plan,
        diaz_negative_retreat=diaz_negative_retreat,
        min_R_noadapt=min_R_noadapt,
    )

    # --------- ELEVATION DISTRIBUTION-INDEPENENT COSTS ----------
    # These are independent of the elevation distribution data so are calculated
    # outside of the below loop

    # SURGE
    # Get different between retreat/protect height and lslr
    rh_diff = (
        RH_heights.sel(at=at, drop=True).isel(return_period=slice(None, -1)) - lslr
    )
    rh_diff_noadapt = lslr_plan_noadapt - lslr

    # rh_diff could be negative if diaz_lslr_plan is true, and in these cases, it is
    # always clipped at 0 for surge damage calculations
    rh_diff = _pos(rh_diff)
    rh_diff_noadapt = _pos(rh_diff_noadapt)

    # calculate fractional storm damage using the coefficients provided by Diaz 2016
    if diaz_storm_calcs:
        tot_landarea = inputs.landarea.sum("elev")
        coefs = inputs.surge_coefs.to_dataset("coef")
        sigma_noadapt = coefs.rsig0 / (
            1 + coefs.rsigA * np.exp(coefs.rsigB * rh_diff_noadapt)
        )
        sigma_r = coefs.rsig0 / (
            1 + coefs.rsigA * np.exp(coefs.rsigB * rh_diff.sel(adapttype="retreat"))
        )
        sigma_p = (coefs.psig0 + coefs.psig0coef * _pos(lslr)) / (
            1 + coefs.psigA * np.exp(coefs.psigB * rh_diff.sel(adapttype="protect"))
        )
        sigma = xr.concat((sigma_r, sigma_p), dim="adapttype")
        surge_cap = sigma / tot_landarea
        surge_pop = surge_cap * inputs.floodmortality
        surge = xr.concat(
            (surge_cap, surge_pop),
            dim=pd.Index(["stormCapital", "stormPopulation"], name="costtype"),
        ).to_dataset("costtype")
        surge_cap_noadapt = sigma_noadapt / tot_landarea
        surge_pop_noadapt = surge_cap_noadapt * inputs.floodmortality
        surge_noadapt = xr.concat(
            (surge_cap_noadapt, surge_pop_noadapt),
            dim=pd.Index(["stormCapital", "stormPopulation"], name="costtype"),
        ).to_dataset("costtype")

        # multiply fractional losses by seg-ir-level total capital and population
        surge["stormCapital"] = surge.stormCapital * inputs.K.sum("elev")
        surge["stormPopulation"] = surge.stormPopulation * inputs.pop.sum("elev")
        surge_noadapt["stormCapital"] = surge_noadapt.stormCapital * inputs.K.sum(
            "elev"
        )
        surge_noadapt[
            "stormPopulation"
        ] = surge_noadapt.stormPopulation * inputs.pop.sum("elev")
    else:
        if surge_lookup is None:
            rh_years = RH_heights.sel(at=at, drop=True)
            min_hts = 0
            max_hts = inputs.surge_height.isel(return_period=-1, drop=True)
            surge_heights_probs = _get_surge_heights_probs(
                min_hts, max_hts, inputs.gumbel_params
            )
            esl = lslr + surge_heights_probs.surge

            surge_noadapt = _calc_storm_damages_no_resilience(
                esl,
                inputs[["K", "pop"]],
                inputs.elev_bounds,
                lslr_plan_noadapt,
                0,
                ddf_i,
                dmf_i,
                ddf_kwargs=ddf_kwargs,
                dmf_kwargs=dmf_kwargs,
                surge_probs=surge_heights_probs.p,
            )

            surge = []
            for R, H in (
                (rh_years.sel(adapttype="retreat", drop=True), 0),
                (0, rh_years.sel(adapttype="protect", drop=True)),
            ):
                surge.append(
                    _calc_storm_damages_no_resilience(
                        esl,
                        inputs[["K", "pop"]],
                        inputs.elev_bounds,
                        R,
                        H,
                        ddf_i,
                        dmf_i,
                        ddf_kwargs=ddf_kwargs,
                        dmf_kwargs=dmf_kwargs,
                        surge_probs=surge_heights_probs.p,
                    )
                )
            surge = xr.concat(
                surge, dim=pd.Index(["retreat", "protect"], name="adapttype")
            )
        else:
            # Interpolate to the SLR and rh_diff values for all scenario/mc/year/iam/ssp
            # lslrs that are too low will result in 0 damages, b/c surge isn't high
            # enough to make it to 0 MSL eleveation, so we fill missings from
            # interpolation with 0's
            surge_noadapt = []
            surge = []
            for seg in inputs.seg.values:
                this_surge_lookup = (
                    surge_lookup.sel(seg=seg)
                    .swap_dims(lslr="lslr_by_seg", rh_diff="rh_diff_by_seg")
                    .reset_coords(drop=True)
                    .frac_losses.rename(lslr_by_seg="lslr", rh_diff_by_seg="rh_diff")
                )
                if this_surge_lookup.sum() == 0:
                    continue
                surge_noadapt.append(
                    this_surge_lookup.sel(adapttype="retreat", drop=True)
                    .interp(
                        lslr=lslr.sel(seg=seg),
                        rh_diff=rh_diff_noadapt.sel(seg=seg),
                        assume_sorted=True,
                        kwargs={"fill_value": 0},
                    )
                    .reset_coords(drop=True)
                    .expand_dims(seg=[seg])
                )

                surge_adapt = []
                for adapttype in this_surge_lookup.adapttype.values:
                    surge_adapt.append(
                        this_surge_lookup.sel(adapttype=adapttype)
                        .interp(
                            lslr=lslr.sel(seg=seg),
                            rh_diff=rh_diff.sel(
                                adapttype=adapttype, seg=seg, drop=True
                            ),
                            assume_sorted=True,
                            kwargs={"fill_value": 0},
                        )
                        .reset_coords(drop=True)
                    )
                surge.append(
                    xr.concat(surge_adapt, dim=this_surge_lookup.adapttype).expand_dims(
                        seg=[seg]
                    )
                )
            if len(surge) == 0:
                surge_noadapt = (
                    rh_diff_noadapt.expand_dims(
                        costtype=this_surge_lookup.costtype
                    ).to_dataset("costtype")
                    * 0
                )
                surge = (
                    rh_diff.expand_dims(costtype=this_surge_lookup.costtype).to_dataset(
                        "costtype"
                    )
                    * 0
                )
            else:
                surge = (
                    xr.concat(surge, dim="seg")
                    .reindex(seg=inputs.seg.values, fill_value=0)
                    .to_dataset("costtype")
                )
                surge_noadapt = (
                    xr.concat(surge_noadapt, dim="seg")
                    .reindex(seg=inputs.seg.values, fill_value=0)
                    .to_dataset("costtype")
                )

            # multiply fractional losses by seg-ir-level total capital and population
            surge["stormCapital"] = surge.stormCapital * inputs.K.sum("elev")
            surge["stormPopulation"] = surge.stormPopulation * inputs.pop.sum("elev")
            surge_noadapt["stormCapital"] = surge_noadapt.stormCapital * inputs.K.sum(
                "elev"
            )
            surge_noadapt[
                "stormPopulation"
            ] = surge_noadapt.stormPopulation * inputs.pop.sum("elev")

    surge = surge.stack(case=["adapttype", "return_period"])
    surge["case"] = (surge.adapttype + surge.return_period.astype(str)).values

    # merge no adapt and retreat/protect, dropping extra protect1 which we do not allow.
    # Also, skip the last return value where we assume full protection
    surge = xr.concat(
        (surge_noadapt.expand_dims(case=["noAdaptation"]), surge), dim="case"
    ).sel(case=["noAdaptation"] + RLIST[:-1] + PLIST[:-1])
    del surge_noadapt

    # monetize deaths
    surge["stormPopulation"] *= inputs.vsl

    # multiply fractional losses by resilience factor
    surge = surge * (1 - inputs.rho)

    # PROTECTION
    # Construction and maintenance costs of seawall. In Diaz 2016, a linear charge was
    # levied at the start of every planning period. This means that the total costs from
    # this linear portion was directly proportional to the number of planning periods.
    # Because we have shortened the planning period, we assume that the lifetime of the
    # seawall foundation that depends on this linear part is 50 years (roughly the
    # length of the planning periods in Diaz 2016). This allows our total seawall
    # construction costs to be roughly independent of the number of planning periods.
    construction_freq = 50
    if diaz_construction_freq:
        construction_freq = tstep_at

    protect_heights = np.array([p.lstrip("protect") for p in PLIST]).astype(int)
    RH_heights_p = RH_heights.sel(
        adapttype="protect", return_period=protect_heights
    ).drop("adapttype")
    RH_heights_p_prev = RH_heights_prev.sel(
        adapttype="protect", return_period=protect_heights
    ).drop("adapttype")

    protection = (
        (
            # construction (one time cost)
            (
                inputs.pcfixedfrac / construction_freq
                + (
                    (1 - inputs.pcfixedfrac)
                    * (RH_heights_p**2 - RH_heights_p_prev**2)
                )
                / tstep_at
            )
            # maintenance (annual cost)
            + inputs.wall_maintcost * RH_heights_p
        )
        * inputs.length
        * inputs.pc
    ).sel(at=at, drop=True)

    # Half of Land rent value lost due to occupation by seawall (rent is
    # assumed to be half of rent of lowest elevation bin, specified in input data
    # creation)
    landrent0 = inputs.landrent
    if "elev" in landrent0.dims:
        landrent0 = landrent0.isel(elev=0, drop=True)

    if diaz_fixed_vars_for_onetime_cost:
        protection = protection + (
            inputs.wall_width2height
            * RH_heights_p
            * landrent0.rename(year="at").sel(at=RH_heights_p.at)
            / 2
            * inputs.length
        ).sel(at=at)
    else:
        protection = protection + (
            inputs.wall_width2height
            * RH_heights_p.sel(at=at, drop=True)
            * landrent0
            / 2
            * inputs.length
        )

    # WETLANDS
    # if protection, all wetlands are lost. For Diaz 2016 inputs, we have more wetlands
    # than land area, so not all wetlands are accounted for in wetlands-by-elev
    if "total_wetland_val" in inputs.data_vars:
        wetland_p = inputs.total_wetland_val
    else:
        wetland_p = (inputs.wetland * inputs.wetlandservice).sum("elev")

    # --------- ELEVATION DISTRIBUTION-DEPENENT COSTS ----------
    def calc_elev_bin_weights(slr, lb_elevs, bin_width):
        """Calculates the fraction of a cell inundated/abandoned given a defined
        slr/retreat height."""
        return _pos(np.minimum(slr - lb_elevs, bin_width)) / bin_width

    # loop over each elevation band to sum up elevation-distribution-dependent costs
    inundation = 0
    relocation_noadapt = 0
    abandonment = 0
    relocation_r = 0
    wetland_r_noadapt = 0

    for e_ix in range(0, len(lb_elev), elev_chunksize):
        this_lb_elev = lb_elev.isel(elev=slice(e_ix, e_ix + elev_chunksize))
        this_elev = this_lb_elev.elev.values

        this_bin_width = bin_width.sel(elev=this_elev)
        this_inputs = inputs.sel(elev=this_elev)
        this_inputs_at_only = this_inputs.rename(year="at").sel(at=RH_heights.at)

        # --------- NO ADAPTATION ----------

        # this is one place where Diaz 2016 employs a forward difference to calculate
        # costs (whereas for retreat/protect scenarios it uses a backwards diff)
        lslr_pos = lslr_plan_noadapt
        if diaz_calc_noadapt_damage_w_lslr:
            lslr_pos = _pos(lslr)
        bin_wt_present = calc_elev_bin_weights(lslr_pos, this_lb_elev, this_bin_width)
        if diaz_forward_diff:
            bin_wt = calc_elev_bin_weights(
                lslr_pos.shift(year=-1).ffill("year"),
                this_lb_elev,
                this_bin_width,
            )
            bin_wt_diff = bin_wt - bin_wt_present
        # else assume no slr before 2000 (to help calculate losses in 2000)
        else:
            bin_wt = bin_wt_present
            bin_wt_diff = bin_wt - calc_elev_bin_weights(
                lslr_pos.shift(year=1, fill_value=0),
                this_lb_elev,
                this_bin_width,
            )
        inundation += (
            # lost land rent
            (bin_wt * this_inputs.landrent * this_inputs.landarea)
            # one-time loss of non-mobile capital
            + bin_wt_diff * (1 - this_inputs.mobcapfrac) * this_inputs.K / tstept
        ).sum("elev")

        # relocation
        relocation_noadapt += (
            bin_wt_diff
            * (
                # one-time relocation of people cost
                this_inputs.reactive_retreat_factor
                * this_inputs.movefactor
                * this_inputs.ypc
                * this_inputs.pop
                # one-time relocation and demolition of capital
                + (
                    this_inputs.capmovefactor * this_inputs.mobcapfrac
                    + this_inputs.democost * (1 - this_inputs.mobcapfrac)
                )
                * this_inputs.K
            )
        ).sum("elev") / tstept

        # --------- RETREAT ----------
        bin_wt = calc_elev_bin_weights(
            RH_heights.sel(adapttype="retreat", drop=True), this_lb_elev, this_bin_width
        )
        bin_wt_diff = _pos(
            bin_wt
            - calc_elev_bin_weights(
                RH_heights_prev.sel(adapttype="retreat", drop=True),
                this_lb_elev,
                this_bin_width,
            )
        )

        # abandonment
        if diaz_fixed_vars_for_onetime_cost:
            abandonment += (
                (
                    # Value of abandoned land following retreat
                    bin_wt * this_inputs_at_only.landrent * this_inputs_at_only.landarea
                    # one-time loss of abandoned capital
                    + bin_wt_diff
                    * (1 - this_inputs_at_only.mobcapfrac)
                    * (1 - this_inputs_at_only.depr)
                    * this_inputs_at_only.K
                    / tstep_at
                )
                .sum("elev")
                .sel(at=at, drop=True)
            )
        else:
            abandonment += (
                # Value of abandoned land following retreat
                (
                    bin_wt.sel(at=at, drop=True)
                    * this_inputs.landrent
                    * this_inputs.landarea
                ).sum("elev")
                # one-time loss of abandoned capital
                + (
                    bin_wt_diff.sel(at=at, drop=True)
                    * (1 - this_inputs.mobcapfrac)
                    * (1 - this_inputs.depr)
                    * this_inputs.K
                    / tstep_at.sel(at=at, drop=True)
                ).sum("elev")
            )

        # relocation
        if diaz_fixed_vars_for_onetime_cost:
            relocation_r += (
                (
                    bin_wt_diff
                    * (
                        # one-time relocation of people
                        this_inputs_at_only.pop
                        * this_inputs_at_only.ypc
                        * this_inputs_at_only.movefactor
                        # one-time relocation and demolition of capital
                        + this_inputs_at_only.K
                        * (
                            this_inputs_at_only.mobcapfrac
                            * this_inputs_at_only.capmovefactor
                            + (1 - this_inputs_at_only.mobcapfrac)
                            * this_inputs_at_only.democost
                        )
                    )
                ).sum("elev")
                / tstep_at
            ).sel(at=at, drop=True)
        else:
            relocation_r += (
                bin_wt_diff.sel(at=at, drop=True)
                * (
                    # one-time relocation of people
                    this_inputs.pop * this_inputs.ypc * this_inputs.movefactor
                    # one-time relocation and demolition of capital
                    + this_inputs.K
                    * (
                        this_inputs.mobcapfrac * this_inputs.capmovefactor
                        + (1 - this_inputs.mobcapfrac) * this_inputs.democost
                    )
                )
            ).sum("elev") / tstep_at.sel(at=at, drop=True)

        # Wetland
        # If no protection, then some fraction of inundated wetlands can accrete if
        # local rate of slr less than max. We define wetlands as covering all land area
        # starting from the bottom of a grid cell. So if there are 5 sqkm of wetlands
        # and 10 sqkm of land area in a grid cell that is half-flooded, all 5 sqkm of
        # wetlands will be lost. If there are more wetlands than land area, the loss
        # is just proportional to inundation percentage. e.g. if that grid cell had 15
        # sqkm of wetlands instead of 5, it would lose 7.5 sqkm when half flooded
        bin_wt = calc_elev_bin_weights(lslr, this_lb_elev, this_bin_width)
        wetland_r_noadapt += (
            this_inputs.wetlandservice
            * np.maximum(
                np.minimum(this_inputs.landarea * bin_wt, this_inputs.wetland),
                this_inputs.wetland * bin_wt,
            )
        ).sum("elev") * np.minimum(1, (localrate / this_inputs.wmaxrate) ** 2)

    # In Diaz 2016, it's the actual costs that are filled for the final year, rather
    # than the LSL heights, for both inundation and relocation
    if diaz_forward_diff:
        yr_mask = inundation.year != inundation.year[-1].item()
        inundation = inundation.where(yr_mask, inundation.isel(year=-2, drop=True))
        relocation_noadapt = relocation_noadapt.where(
            yr_mask, relocation_noadapt.isel(year=-2, drop=True)
        )

    # --------- Aggregate costs ------------

    def caseify(ds, prefix):
        ds = ds.rename(return_period="case")
        ds["case"] = np.char.add(prefix, ds.case.astype(str))
        return ds

    abandonment = caseify(abandonment, "retreat")
    relocation_r = caseify(relocation_r, "retreat")
    protection = caseify(protection, "protect")

    # No Adaptation
    noadapt_cost = (
        xr.merge(
            (
                surge.sel(case="noAdaptation", drop=True),
                xr.Dataset(
                    {
                        "inundation": inundation,
                        "relocation": relocation_noadapt,
                        "wetland": wetland_r_noadapt,
                        "protection": xr.zeros_like(inundation),
                    }
                ),
            )
        )
        .to_array("costtype")
        .expand_dims(case=["noAdaptation"])
    )
    del inundation, relocation_noadapt

    # Retreat
    retreat_cost = xr.merge(
        (
            # this fills retreat10000 with a 0 surge damage
            surge.reindex(case=RLIST, fill_value=0),
            xr.Dataset(
                {
                    "inundation": abandonment,
                    "relocation": relocation_r,
                    "wetland": wetland_r_noadapt,
                    "protection": xr.zeros_like(abandonment),
                }
            ),
        )
    ).to_array("costtype")
    del abandonment, relocation_r, wetland_r_noadapt

    # Protect
    protect_cost = xr.merge(
        (
            # this fills protect10000 with a 0 surge damage
            surge.reindex(case=PLIST, fill_value=0),
            xr.Dataset(
                {
                    "inundation": xr.zeros_like(protection),
                    "relocation": xr.zeros_like(protection),
                    "wetland": wetland_p,
                    "protection": protection,
                }
            ),
        )
    ).to_array("costtype")
    del protection, wetland_p

    # Aggregate costs for all cases and mutliply by length of each timestep
    out = xr.concat(
        (noadapt_cost, retreat_cost, protect_cost),
        dim="case",
    )
    del noadapt_cost, retreat_cost, protect_cost

    # Diaz 2016 assigns 0 cost to all with-adaptation scenarios in the first year, with
    # the exception of wetland costs
    if diaz_zero_costs_in_first_year:
        out.where(
            (out.year > out.year[0])
            | (out.case == "noAdaptation")
            | (out.costtype == "wetland"),
            0,
        )

    if return_year0_hts:
        RH_heights0 = RH_heights.isel(at=0, drop=True)
        out = [
            out,
            xr.concat(
                (
                    lslr_plan_noadapt.isel(year=0, drop=True).expand_dims(
                        case=["noAdaptation"]
                    ),
                    caseify(RH_heights0.sel(adapttype="retreat", drop=True), "retreat"),
                    caseify(RH_heights0.sel(adapttype="protect", drop=True), "protect"),
                ),
                dim="case",
            ).sel(case=out.case),
        ]
    if return_RH_heights:
        RH_heights = RH_heights.stack(case=["adapttype", "return_period"])
        RH_heights["case"] = RH_heights.adapttype + RH_heights.return_period.astype(str)
        RH_heights = xr.concat(
            (
                lslr_plan_noadapt.expand_dims(case=["noAdaptation"]),
                RH_heights.sel(at=at, drop=True),
            ),
            dim="case",
        )
        if isinstance(out, list):
            out.append(RH_heights)
        else:
            out = [out, RH_heights]

    return out


def select_optimal_case(all_case_cost_path, region, seg_regions, region_var="seg_adm"):
    """Calculate the least-cost adaptation path for a given `region`, which is nested
    within a given coastal segment. All regions within a segment must take the same
    adaptation strategy.

    Parameters
    ----------
    all_case_cost_path : Path-like
        Path to Zarr store that contains ``costs`` and ``npv`` variables for each
        adaptation choice for all regions.
    region : str
        Name of region that you will calculate optimal case for
    seg_regions : list of str
        Names of all regions within this segment. NPV across all regions will be summed
        to calculate the segment-level least-cost adaptation choice.
    region_var : str, default "seg_adm"
        Name of dimension corresponding to region name.

    Returns
    -------
    :py:class:`xarray.Dataset`
        Same as the Dataset stored at `all_case_cost_path` but with only one ``case``,
        ``optimalfixed``, which represents the optimal adaptation choice for this region
        for each socioeconomic and SLR trajectory.
    """

    opt_case = (
        xr.open_zarr(all_case_cost_path, chunks=None)
        .npv.sel({region_var: seg_regions})
        .drop_sel(case="optimalfixed")
        .sum(region_var)
        .idxmin("case")
    )
    opt_case_ser = opt_case.to_series()
    opt_val = (
        pd.Series(
            pd.Series(CASE_DICT).loc[opt_case_ser].values,
            index=opt_case_ser.index,
        )
        .to_xarray()
        .astype("uint8")
    )

    out = (
        xr.open_zarr(all_case_cost_path, chunks=None)[["costs", "npv"]]
        .sel({region_var: [region]})
        .sel(case=opt_case, drop=True)
        .expand_dims(case=["optimalfixed"])
    )
    out["optimal_case"] = opt_val.expand_dims({region_var: [region]})
    return out
